@@ -23,7 +23,7 @@
 #ifdef HAVE_STDINT_H
 #include <stdint.h>     /* For uint32_t */
 #endif
-#ifndef _WIN32
+#if !defined(WIN32) && !defined(WIN64)
 #include <netdb.h>  /* For gethostbyname */
 #include <netinet/in.h> /* For ntohl */
 #else
@@ -50,6 +50,7 @@ int main () {
   char ipAddress[30];
   char expectedCountry[3];
   char expectedCountry3[4];
+  const char * time_zone;
 
 	gi = GeoIP_open("../data/GeoIPRegion.dat", GEOIP_MEMORY_CACHE);
 
@@ -77,10 +78,13 @@ int main () {
 		printf("ip = %s\n",ipAddress);
 
 		gir = GeoIP_region_by_name (gi, ipAddress);
+		time_zone = GeoIP_time_zone_by_country_and_region(gir->country_code, gir->region);
 		if (gir != NULL) {
-			printf("%s, %s\n",
-						 gir->country_code,
-						 (!gir->region[0]) ? "N/A" : gir->region );
+			printf("%s, %s, %s, %s\n",
+						gir->country_code,
+						(!gir->region[0]) ? "N/A" : gir->region,
+						GeoIP_region_name_by_code(gir->country_code, gir->region),
+						time_zone);
 		} else {
 			printf("NULL!\n");
 		}
